import asyncio

import httpcore
import httpx
import pytest

from sanic_testing.testing import SanicTestClient

from sanic import Sanic
from sanic.response import text


class DelayableHTTPConnection(httpcore._async.connection.AsyncHTTPConnection):
    async def arequest(self, *args, **kwargs):
        await asyncio.sleep(2)
        return await super().arequest(*args, **kwargs)

    async def _open_socket(self, *args, **kwargs):
        retval = await super()._open_socket(*args, **kwargs)
        if self._request_delay:
            await asyncio.sleep(self._request_delay)
        return retval


class DelayableSanicConnectionPool(httpcore.AsyncConnectionPool):
    def __init__(self, request_delay=None, *args, **kwargs):
        self._request_delay = request_delay
        super().__init__(*args, **kwargs)

    async def _add_to_pool(self, connection, timeout):
        connection.__class__ = DelayableHTTPConnection
        connection._request_delay = self._request_delay
        await super()._add_to_pool(connection, timeout)


class DelayableSanicSession(httpx.AsyncClient):
    def __init__(self, request_delay=None, *args, **kwargs) -> None:
        transport = DelayableSanicConnectionPool(request_delay=request_delay)
        super().__init__(transport=transport, *args, **kwargs)


class DelayableSanicTestClient(SanicTestClient):
    def __init__(self, app, request_delay=None):
        super().__init__(app)
        self._request_delay = request_delay
        self._loop = None

    def get_new_session(self):
        return DelayableSanicSession(request_delay=self._request_delay)


@pytest.fixture
def request_no_timeout_app():
    app = Sanic("test_request_no_timeout")
    app.config.REQUEST_TIMEOUT = 0.6

    @app.route("/1")
    async def handler2(request):
        return text("OK")

    return app


@pytest.fixture
def request_timeout_default_app():
    app = Sanic("test_request_timeout_default")
    app.config.REQUEST_TIMEOUT = 0.6

    @app.route("/1")
    async def handler1(request):
        return text("OK")

    @app.websocket("/ws1")
    async def ws_handler1(request, ws):
        await ws.send("OK")

    return app


def test_default_server_error_request_timeout(request_timeout_default_app):
    client = DelayableSanicTestClient(request_timeout_default_app, 2)
    _, response = client.get("/1")
    assert response.status == 408
    assert "Request Timeout" in response.text


def test_default_server_error_request_dont_timeout(request_no_timeout_app):
    client = DelayableSanicTestClient(request_no_timeout_app, 0.2)
    _, response = client.get("/1")
    assert response.status == 200
    assert response.text == "OK"


def test_default_server_error_websocket_request_timeout(
    request_timeout_default_app,
):

    headers = {
        "Upgrade": "websocket",
        "Connection": "upgrade",
        "Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ==",
        "Sec-WebSocket-Version": "13",
    }

    client = DelayableSanicTestClient(request_timeout_default_app, 2)
    _, response = client.get("/ws1", headers=headers)

    assert response.status == 408
    assert "Request Timeout" in response.text



import asyncio
import logging
import time

from collections import Counter
from multiprocessing import Process

import httpx


PORT = 42101


def test_no_exceptions_when_cancel_pending_request(app, caplog):
    app.config.GRACEFUL_SHUTDOWN_TIMEOUT = 1

    @app.get("/")
    async def handler(request):
        await asyncio.sleep(5)

    @app.after_server_start
    def shutdown(app, _):
        time.sleep(0.2)
        app.stop()

    def ping():
        time.sleep(0.1)
        response = httpx.get("http://127.0.0.1:8000")
        print(response.status_code)

    p = Process(target=ping)
    p.start()

    with caplog.at_level(logging.INFO):
        app.run()

    p.kill()

    counter = Counter([r[1] for r in caplog.record_tuples])

    assert counter[logging.INFO] == 5
    assert logging.ERROR not in counter
    assert (
        caplog.record_tuples[3][2]
        == "Request: GET http://127.0.0.1:8000/ stopped. Transport is closed."
    )



import logging

from sanic.signals import RESERVED_NAMESPACES
from sanic.touchup import TouchUp


def test_touchup_methods(app):
    assert len(TouchUp._registry) == 9


async def test_ode_removes_dispatch_events(app, caplog):
    with caplog.at_level(logging.DEBUG, logger="sanic.root"):
        await app._startup()
    logs = caplog.record_tuples

    for signal in RESERVED_NAMESPACES["http"]:
        assert (
            "sanic.root",
            logging.DEBUG,
            f"Disabling event: {signal}",
        ) in logs



import asyncio
import json as stdjson

from collections import namedtuple
from textwrap import dedent
from typing import AnyStr

import pytest

from sanic_testing.reusable import ReusableClient

from sanic import json, text
from sanic.app import Sanic


PORT = 1234


class RawClient:
    CRLF = b"\r\n"

    def __init__(self, host: str, port: int):
        self.reader = None
        self.writer = None
        self.host = host
        self.port = port

    async def connect(self):
        self.reader, self.writer = await asyncio.open_connection(
            self.host, self.port
        )

    async def close(self):
        self.writer.close()
        await self.writer.wait_closed()

    async def send(self, message: AnyStr):
        if isinstance(message, str):
            msg = self._clean(message).encode("utf-8")
        else:
            msg = message
        await self._send(msg)

    async def _send(self, message: bytes):
        if not self.writer:
            raise Exception("No open write stream")
        self.writer.write(message)

    async def recv(self, nbytes: int = -1) -> bytes:
        if not self.reader:
            raise Exception("No open read stream")
        return await self.reader.read(nbytes)

    def _clean(self, message: str) -> str:
        return (
            dedent(message)
            .lstrip("\n")
            .replace("\n", self.CRLF.decode("utf-8"))
        )


@pytest.fixture
def test_app(app: Sanic):
    app.config.KEEP_ALIVE_TIMEOUT = 1

    @app.get("/")
    async def base_handler(request):
        return text("111122223333444455556666777788889999")

    @app.post("/upload", stream=True)
    async def upload_handler(request):
        data = [part.decode("utf-8") async for part in request.stream]
        return json(data)

    return app


@pytest.fixture
def runner(test_app):
    client = ReusableClient(test_app, port=PORT)
    client.run()
    yield client
    client.stop()


@pytest.fixture
def client(runner):
    client = namedtuple("Client", ("raw", "send", "recv"))

    raw = RawClient(runner.host, runner.port)
    runner._run(raw.connect())

    def send(msg):
        nonlocal runner
        nonlocal raw
        runner._run(raw.send(msg))

    def recv(**kwargs):
        nonlocal runner
        nonlocal raw
        method = raw.recv_until if "until" in kwargs else raw.recv
        return runner._run(method(**kwargs))

    yield client(raw, send, recv)

    runner._run(raw.close())


def test_full_message(client):
    client.send(
        """
        GET / HTTP/1.1
        host: localhost:7777

        """
    )
    response = client.recv()
    assert len(response) == 140
    assert b"200 OK" in response


def test_transfer_chunked(client):
    client.send(
        """
        POST /upload HTTP/1.1
        transfer-encoding: chunked

        """
    )
    client.send(b"3\r\nfoo\r\n")
    client.send(b"3\r\nbar\r\n")
    client.send(b"0\r\n\r\n")
    response = client.recv()
    _, body = response.rsplit(b"\r\n\r\n", 1)
    data = stdjson.loads(body)

    assert data == ["foo", "bar"]



import asyncio
import logging

from sanic_testing.testing import PORT

from sanic.config import BASE_LOGO


def test_logo_base(app, run_startup):
    logs = run_startup(app)

    assert logs[0][1] == logging.DEBUG
    assert logs[0][2] == BASE_LOGO


def test_logo_false(app, caplog, run_startup):
    app.config.LOGO = False

    logs = run_startup(app)

    banner, port = logs[0][2].rsplit(":", 1)
    assert logs[0][1] == logging.INFO
    assert banner == "Goin' Fast @ http://127.0.0.1"
    assert int(port) > 0


def test_logo_true(app, run_startup):
    app.config.LOGO = True

    logs = run_startup(app)

    assert logs[0][1] == logging.DEBUG
    assert logs[0][2] == BASE_LOGO


def test_logo_custom(app, run_startup):
    app.config.LOGO = "My Custom Logo"

    logs = run_startup(app)

    assert logs[0][1] == logging.DEBUG
    assert logs[0][2] == "My Custom Logo"
